@@ -29,12 +29,13 @@
 #include "drkonqi_globals.h"
 #include "applicationdetailsexamples.h"
 
-#include <QtGui/QLabel>
-#include <QtGui/QCheckBox>
-#include <QtGui/QToolTip>
+#include <QLabel>
+#include <QCheckBox>
+#include <QToolTip>
 
 #include <KToolInvocation>
 #include <KIcon>
+#include <KUrl>
 #include <KMessageBox>
 
 //BEGIN IntroductionPage
