@@ -3,6 +3,7 @@
 
     Copyright (C) 2002 Waldo Bastian (bastian@kde.org)
     Copyright (C) 2005 David Faure (faure@kde.org)
+    Copyright (C) 2012 Dawit Alemayehu (adawit@kde.org)
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -24,9 +25,7 @@
 
 #include "kpasswdserver.h"
 
-#include <time.h>
-
-#include <QtCore/QTimer>
+#include "kpasswdserveradaptor.h"
 
 #include <kapplication.h>
 #include <klocale.h>
@@ -39,7 +38,10 @@
 #include <kpluginfactory.h>
 #include <kpluginloader.h>
 
-#include "kpasswdserveradaptor.h"
+#include <QtCore/QTimer>
+
+#include <ctime>
+
 
 K_PLUGIN_FACTORY(KPasswdServerFactory,
                  registerPlugin<KPasswdServer>();
@@ -76,24 +78,29 @@ KPasswdServer::KPasswdServer(QObject* parent, const QList<QVariant>&)
  : KDEDModule(parent)
 {
     KIO::AuthInfo::registerMetaTypes();
+
     m_seqNr = 0;
     m_wallet = 0;
+    m_walletDisabled = false;
 
     KPasswdServerAdaptor *adaptor = new KPasswdServerAdaptor(this);
     // register separately from kded
     QDBusConnection::sessionBus().registerService("org.kde.kpasswdserver");
     // connect signals to the adaptor
     connect(this,
-            SIGNAL(checkAuthInfoAsyncResult(qlonglong, qlonglong, const KIO::AuthInfo &)),
+            SIGNAL(checkAuthInfoAsyncResult(qlonglong,qlonglong,KIO::AuthInfo)),
             adaptor,
-            SIGNAL(checkAuthInfoAsyncResult(qlonglong, qlonglong, const KIO::AuthInfo &)));
+            SIGNAL(checkAuthInfoAsyncResult(qlonglong,qlonglong,KIO::AuthInfo)));
     connect(this,
-            SIGNAL(queryAuthInfoAsyncResult(qlonglong, qlonglong, const KIO::AuthInfo &)),
+            SIGNAL(queryAuthInfoAsyncResult(qlonglong,qlonglong,KIO::AuthInfo)),
             adaptor,
-            SIGNAL(queryAuthInfoAsyncResult(qlonglong, qlonglong, const KIO::AuthInfo &)));
+            SIGNAL(queryAuthInfoAsyncResult(qlonglong,qlonglong,KIO::AuthInfo)));
 
     connect(this, SIGNAL(windowUnregistered(qlonglong)),
             this, SLOT(removeAuthForWindowId(qlonglong)));
+
+    connect(KWindowSystem::self(), SIGNAL(windowRemoved(WId)),
+            this, SLOT(windowRemoved(WId)));
 }
 
 KPasswdServer::~KPasswdServer()
@@ -103,6 +110,8 @@ KPasswdServer::~KPasswdServer()
     qDeleteAll(m_authPending);
     qDeleteAll(m_authWait);
     qDeleteAll(m_authDict);
+    qDeleteAll(m_authInProgress);
+    qDeleteAll(m_authRetryInProgress);
     delete m_wallet;
 }
 
@@ -198,15 +207,15 @@ static bool readFromWallet( KWallet::Wallet* wallet, const QString& key, const Q
 
 bool KPasswdServer::hasPendingQuery(const QString &key, const KIO::AuthInfo &info)
 {
-    QString path2 = info.url.directory(KUrl::AppendTrailingSlash | KUrl::ObeyTrailingSlash);
+    const QString path2 (info.url.directory(KUrl::AppendTrailingSlash | KUrl::ObeyTrailingSlash));
     Q_FOREACH(const Request *request, m_authPending) {
         if (request->key != key) {
             continue;
         }
 
         if (info.verifyPath) {
-            QString path1 = request->info.url.directory(KUrl::AppendTrailingSlash |
-                                                        KUrl::ObeyTrailingSlash);
+            const QString path1 (request->info.url.directory(KUrl::AppendTrailingSlash |
+                                                             KUrl::ObeyTrailingSlash));
             if (!path2.startsWith(path1)) {
                 continue;
             }
@@ -224,14 +233,13 @@ KPasswdServer::checkAuthInfo(const QByteArray &data, qlonglong windowId, qlonglo
     KIO::AuthInfo info;
     QDataStream stream(data);
     stream >> info;
-    kDebug(debugArea()) << "User =" << info.username << ", WindowId =" << windowId << endl;
     if (usertime != 0) {
         kapp->updateUserTimestamp(usertime);
     }
 
     // if the check depends on a pending query, delay it
     // until that query is finished.
-    QString key = createCacheKey(info);
+    const QString key (createCacheKey(info));
     if (hasPendingQuery(key, info)) {
         setDelayedReply(true);
         Request *pendingCheck = new Request;
@@ -245,6 +253,7 @@ KPasswdServer::checkAuthInfo(const QByteArray &data, qlonglong windowId, qlonglo
         return data;             // return value will be ignored
     }
 
+    // kDebug(debugArea()) << "key =" << key << "user =" << info.username << "windowId =" << windowId;
     const AuthInfoContainer *result = findAuthInfoItem(key, info);
     if (!result || result->isCanceled)
     {
@@ -267,8 +276,9 @@ KPasswdServer::checkAuthInfo(const QByteArray &data, qlonglong windowId, qlonglo
             info.setModified(false);
         }
     } else {
+        kDebug(debugArea()) << "Found cached authentication for" << key;
         updateAuthExpire(key, result, windowId, false);
-        info = copyAuthInfo(result);
+        copyAuthInfo(result, info);
     }
 
     QByteArray data2;
@@ -294,7 +304,7 @@ qlonglong KPasswdServer::checkAuthInfoAsync(KIO::AuthInfo info, qlonglong window
 
     // if the check depends on a pending query, delay it
     // until that query is finished.
-    QString key = createCacheKey(info);
+    const QString key (createCacheKey(info));
     if (hasPendingQuery(key, info)) {
         Request *pendingCheck = new Request;
         pendingCheck->isAsync = true;
@@ -327,8 +337,9 @@ qlonglong KPasswdServer::checkAuthInfoAsync(KIO::AuthInfo info, qlonglong window
             info.setModified(false);
         }
     } else {
+        // kDebug(debugArea()) << "Found cached authentication for" << key;
         updateAuthExpire(key, result, windowId, false);
-        info = copyAuthInfo(result);
+        copyAuthInfo(result, info);
     }
 
     emit checkAuthInfoAsyncResult(requestId, m_seqNr, info);
@@ -342,15 +353,18 @@ KPasswdServer::queryAuthInfo(const QByteArray &data, const QString &errorMsg,
     KIO::AuthInfo info;
     QDataStream stream(data);
     stream >> info;
-    kDebug(debugArea()) << "User =" << info.username << ", Message= " << info.prompt
-                << ", WindowId =" << windowId << endl;
-    if ( !info.password.isEmpty() ) // should we really allow the caller to pre-fill the password?
+
+    kDebug(debugArea()) << "User =" << info.username << ", WindowId =" << windowId
+                        << "seqNr =" << seqNr << ", errorMsg =" << errorMsg;
+
+    if ( !info.password.isEmpty() ) { // should we really allow the caller to pre-fill the password?
         kDebug(debugArea()) << "password was set by caller";
+    }
     if (usertime != 0) {
         kapp->updateUserTimestamp(usertime);
     }
 
-    QString key = createCacheKey(info);
+    const QString key (createCacheKey(info));
     Request *request = new Request;
     setDelayedReply(true);
     request->isAsync = false;
@@ -381,8 +395,9 @@ qlonglong
 KPasswdServer::queryAuthInfoAsync(const KIO::AuthInfo &info, const QString &errorMsg,
                                   qlonglong windowId, qlonglong seqNr, qlonglong usertime)
 {
-    kDebug(debugArea()) << "User =" << info.username << ", Message= " << info.prompt
-                << ", WindowId =" << windowId << endl;
+    kDebug(debugArea()) << "User =" << info.username << ", WindowId =" << windowId
+                        << "seqNr =" << seqNr << ", errorMsg =" << errorMsg;
+
     if (!info.password.isEmpty()) {
         kDebug(debugArea()) << "password was set by caller";
     }
@@ -390,7 +405,7 @@ KPasswdServer::queryAuthInfoAsync(const KIO::AuthInfo &info, const QString &erro
         kapp->updateUserTimestamp(usertime);
     }
 
-    QString key = createCacheKey(info);
+    const QString key (createCacheKey(info));
     Request *request = new Request;
     request->isAsync = true;
     request->requestId = getRequestId();
@@ -417,13 +432,12 @@ KPasswdServer::queryAuthInfoAsync(const KIO::AuthInfo &info, const QString &erro
 void
 KPasswdServer::addAuthInfo(const KIO::AuthInfo &info, qlonglong windowId)
 {
-    kDebug(debugArea()) << "User =" << info.username << ", RealmValue= " << info.realmValue
-                        << ", WindowId = " << windowId << endl;
-    const QString key = createCacheKey(info);
+    kDebug(debugArea()) << "User =" << info.username << ", Realm =" << info.realmValue << ", WindowId =" << windowId;
+    const QString key (createCacheKey(info));
 
     m_seqNr++;
 
-    if (openWallet(windowId) && storeInWallet(m_wallet, key, info)) {
+    if (!m_walletDisabled && openWallet(windowId) && storeInWallet(m_wallet, key, info)) {
         // Since storing the password in the wallet succeeded, make sure the
         // password information is stored in memory only for the duration the
         // windows associated with it are still around.
@@ -442,15 +456,13 @@ KPasswdServer::addAuthInfo(const QByteArray &data, qlonglong windowId)
     KIO::AuthInfo info;
     QDataStream stream(data);
     stream >> info;
-    kDebug(debugArea()) << "User =" << info.username << ", RealmValue= " << info.realmValue
-                << ", WindowId = " << windowId << endl;
     addAuthInfo(info, windowId);
 }
 
 void
 KPasswdServer::removeAuthInfo(const QString& host, const QString& protocol, const QString& user)
 {
-    kDebug(debugArea()) << "Wanted" << protocol << host << user;
+    kDebug(debugArea()) << protocol << host << user;
 
     QHashIterator< QString, AuthInfoContainerList* > dictIterator(m_authDict);
     while (dictIterator.hasNext())
@@ -478,7 +490,7 @@ KPasswdServer::removeAuthInfo(const QString& host, const QString& protocol, cons
 }
 
 bool
-KPasswdServer::openWallet( int windowId )
+KPasswdServer::openWallet( qlonglong windowId )
 {
     if ( m_wallet && !m_wallet->isOpen() ) { // forced closed
         delete m_wallet;
@@ -486,7 +498,7 @@ KPasswdServer::openWallet( int windowId )
     }
     if ( !m_wallet )
         m_wallet = KWallet::Wallet::openWallet(
-            KWallet::Wallet::NetworkWallet(), (WId) windowId );
+            KWallet::Wallet::NetworkWallet(), static_cast<WId>(windowId));
     return m_wallet != 0;
 }
 
@@ -511,11 +523,17 @@ KPasswdServer::processRequest()
     m_authPrompted.append(request->key);
 
     KIO::AuthInfo &info = request->info;
-    bool bypassCacheAndKWallet = info.getExtraField(AUTHINFO_EXTRAFIELD_BYPASS_CACHE_AND_KWALLET).toBool();
-    bool skipAutoCaching = info.getExtraField(AUTHINFO_EXTRAFIELD_SKIP_CACHING_ON_QUERY).toBool();
 
-    kDebug(debugArea()) << "User =" << info.username << ", Message =" << info.prompt;
+    // NOTE: If info.username is empty and info.url.user() is not, set
+    // info.username to info.url.user() to ensure proper caching. See
+    // note passwordDialogDone.
+    if (info.username.isEmpty() && !info.url.user().isEmpty()) {
+        info.username = info.url.user();
+    }
+    const bool bypassCacheAndKWallet = info.getExtraField(AUTHINFO_EXTRAFIELD_BYPASS_CACHE_AND_KWALLET).toBool();
+
     const AuthInfoContainer *result = findAuthInfoItem(request->key, request->info);
+    kDebug(debugArea()) << "key=" << request->key << ", user=" << info.username << "seqNr: request=" << request->seqNr << ", result=" << (result ? result->seqNr : -1);
 
     if (!bypassCacheAndKWallet && result && (request->seqNr < result->seqNr))
     {
@@ -527,131 +545,53 @@ KPasswdServer::processRequest()
         else
         {
            updateAuthExpire(request->key, result, request->windowId, false);
-           info = copyAuthInfo(result);
+           copyAuthInfo(result, info);
         }
     }
     else
     {
         m_seqNr++;
-        bool askPw = request->prompt;
         if (result && !request->errorMsg.isEmpty())
         {
-           QString prompt = request->errorMsg + "  ";
-           prompt += i18n("Do you want to retry?");
-           int dlgResult = KMessageBox::warningContinueCancel(0, prompt,
-                           i18n("Authentication"), KGuiItem(i18n("Retry")));
-           if (dlgResult != KMessageBox::Continue)
-              askPw = false;
+            QString prompt (request->errorMsg.trimmed());
+            prompt += QLatin1Char('\n');
+            prompt += i18n("Do you want to retry?");
+
+            KDialog* dlg = new KDialog(0, Qt::Dialog);
+            connect(dlg, SIGNAL(finished(int)), this, SLOT(retryDialogDone(int)));
+            connect(this, SIGNAL(destroyed(QObject*)), dlg, SLOT(deleteLater()));
+            dlg->setPlainCaption(i18n("Retry Authorization"));
+            dlg->setWindowIcon(KIcon("dialog-password"));
+            dlg->setButtons(KDialog::Yes | KDialog::No);
+            dlg->setObjectName("warningOKCancel");
+            KGuiItem buttonContinue (i18nc("@action:button filter-continue", "Retry"));
+            dlg->setButtonGuiItem(KDialog::Yes, buttonContinue);
+            dlg->setButtonGuiItem(KDialog::No, KStandardGuiItem::cancel());
+            dlg->setDefaultButton(KDialog::Yes);
+            dlg->setEscapeButton(KDialog::No);
+
+            KMessageBox::createKMessageBox(dlg, QMessageBox::Warning, prompt,
+                                           QStringList(), QString(), 0L,
+                                           (KMessageBox::Notify | KMessageBox::NoExec));
+
+        #ifndef Q_WS_WIN
+            KWindowSystem::setMainWindow(dlg, request->windowId);
+        #else
+            KWindowSystem::setMainWindow(dlg, (HWND)(long)request->windowId);
+        #endif
+
+            kDebug(debugArea()) << "Calling open on retry dialog" << dlg;
+            m_authRetryInProgress.insert(dlg, request.take());
+            dlg->open();
+            return;
         }
 
-        int dlgResult = QDialog::Rejected;
-        if (askPw)
+        if (request->prompt)
         {
-            QString username = info.username;
-            QString password = info.password;
-            bool hasWalletData = false;
-            QMap<QString, QString> knownLogins;
-
-            if ( !bypassCacheAndKWallet
-                && ( username.isEmpty() || password.isEmpty() )
-                && !KWallet::Wallet::keyDoesNotExist(KWallet::Wallet::NetworkWallet(), KWallet::Wallet::PasswordFolder(), makeWalletKey( request->key, info.realmValue )) )
-            {
-                // no login+pass provided, check if kwallet has one
-                if ( openWallet( request->windowId ) )
-                    hasWalletData = readFromWallet( m_wallet, request->key, info.realmValue, username, password, info.readOnly, knownLogins );
-            }
-
-            // assemble dialog-flags
-            KPasswordDialog::KPasswordDialogFlags dialogFlags;
-
-            if (info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).isValid())
-            {
-                dialogFlags |= KPasswordDialog::ShowDomainLine;
-                if (info.getExtraFieldFlags(AUTHINFO_EXTRAFIELD_DOMAIN) & KIO::AuthInfo::ExtraFieldReadOnly)
-                {
-                    dialogFlags |= KPasswordDialog::DomainReadOnly;
-                }
-            }
-
-            if (info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).isValid())
-            {
-                dialogFlags |= KPasswordDialog::ShowAnonymousLoginCheckBox;
-            }
-
-            if (!info.getExtraField(AUTHINFO_EXTRAFIELD_HIDE_USERNAME_INPUT).toBool())
-            {
-                dialogFlags |= KPasswordDialog::ShowUsernameLine;
-            }
-
-            // If wallet is not enabled and the caller explicitly request for it,
-            // do not show the keep password checkbox.
-            if (info.keepPassword && KWallet::Wallet::isEnabled())
-                dialogFlags |= KPasswordDialog::ShowKeepPassword;
-
-            // instantiate dialog
-            KPasswordDialog dlg( 0l, dialogFlags) ;
-            dlg.setPrompt(info.prompt);
-            dlg.setUsername(username);
-            if (info.caption.isEmpty())
-               dlg.setPlainCaption( i18n("Authorization Dialog") );
-            else
-               dlg.setPlainCaption( info.caption );
-
-            if ( !info.comment.isEmpty() )
-               dlg.addCommentLine( info.commentLabel, info.comment );
-
-            if ( !password.isEmpty() )
-               dlg.setPassword( password );
-
-            if (info.readOnly)
-               dlg.setUsernameReadOnly( true );
-            else
-               dlg.setKnownLogins( knownLogins );
-
-            if (hasWalletData)
-                dlg.setKeepPassword( true );
-
-            if (info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).isValid ())
-                dlg.setDomain(info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).toString());
-
-            if (info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).isValid () && password.isEmpty() && username.isEmpty())
-                dlg.setAnonymousMode(info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).toBool());
-
-#ifndef Q_WS_WIN
-            KWindowSystem::setMainWindow(&dlg, request->windowId);
-#else
-            KWindowSystem::setMainWindow(&dlg, (HWND)(long)request->windowId);
-#endif
-
-            kDebug(debugArea()) << "Calling exec on" << &dlg;
-            dlgResult = dlg.exec();
-            kDebug(debugArea()) << "exec returned with" << dlgResult;
-
-            if (dlgResult == QDialog::Accepted)
-            {
-               info.username = dlg.username();
-               info.password = dlg.password();
-               info.keepPassword = dlg.keepPassword();
-
-               if (info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).isValid ())
-                   info.setExtraField(AUTHINFO_EXTRAFIELD_DOMAIN, dlg.domain());
-               if (info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).isValid ())
-                   info.setExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS, dlg.anonymousMode());
-
-               // When the user checks "keep password", that means:
-               // * if the wallet is enabled, store it there for long-term, and in kpasswdserver
-               // only for the duration of the window (#92928)
-               // * otherwise store in kpasswdserver for the duration of the KDE session.
-               if ( !bypassCacheAndKWallet && !skipAutoCaching && info.keepPassword ) {
-                   if ( openWallet( request->windowId ) ) {
-                       if ( storeInWallet( m_wallet, request->key, info ) )
-                           // password is in wallet, don't keep it in memory after window is closed
-                           info.keepPassword = false;
-                   }
-               }
-            }
+            showPasswordDialog(request.take());
+            return;
         }
-        if ( dlgResult != QDialog::Accepted )
+        else
         {
             if (!bypassCacheAndKWallet && request->prompt)
             {
@@ -659,70 +599,9 @@ KPasswdServer::processRequest()
             }
             info.setModified( false );
         }
-        else
-        {
-            if (!bypassCacheAndKWallet)
-            {
-                addAuthInfoItem(request->key, info, request->windowId, m_seqNr, false);
-            }
-            info.setModified( true );
-        }
     }
 
-    if (request->isAsync) {
-        emit queryAuthInfoAsyncResult(request->requestId, m_seqNr, info);
-    } else {
-        QByteArray replyData;
-        QDataStream stream2(&replyData, QIODevice::WriteOnly);
-        stream2 << info;
-        QDBusConnection::sessionBus().send(request->transaction.createReply(QVariantList() << replyData << m_seqNr));
-    }
-
-    // Check all requests in the wait queue.
-    Request *waitRequest;
-    QMutableListIterator<Request*> it(m_authWait);
-    while (it.hasNext()) {
-        waitRequest = it.next();
-
-        if (!hasPendingQuery(waitRequest->key, waitRequest->info))
-        {
-            const AuthInfoContainer *result = findAuthInfoItem(waitRequest->key,
-                                                               waitRequest->info);
-            QByteArray replyData;
-
-            QDataStream stream2(&replyData, QIODevice::WriteOnly);
-
-            KIO::AuthInfo rcinfo;
-            if (!result || result->isCanceled)
-            {
-                waitRequest->info.setModified(false);
-                stream2 << waitRequest->info;
-            }
-            else
-            {
-                updateAuthExpire(waitRequest->key, result, waitRequest->windowId, false);
-                rcinfo = copyAuthInfo(result);
-                stream2 << rcinfo;
-            }
-
-            if (waitRequest->isAsync) {
-                emit checkAuthInfoAsyncResult(waitRequest->requestId, m_seqNr, rcinfo);
-            } else {
-                QDBusConnection::sessionBus().send(waitRequest->transaction.createReply(QVariantList() << replyData << m_seqNr));
-            }
-
-            delete waitRequest;
-            it.remove();
-        }
-    }
-
-    // reallow processing for the specific window.
-    m_authPrompted.removeAll(windowIdStr);
-    m_authPrompted.removeAll(request->key);
-
-    if (m_authPending.count())
-       QTimer::singleShot(0, this, SLOT(processRequest()));
-
+    sendResponse(request.data());
 }
 
 QString KPasswdServer::createCacheKey( const KIO::AuthInfo &info )
@@ -752,48 +631,45 @@ QString KPasswdServer::createCacheKey( const KIO::AuthInfo &info )
     return key;
 }
 
-KIO::AuthInfo
-KPasswdServer::copyAuthInfo(const AuthInfoContainer *i)
+void KPasswdServer::copyAuthInfo(const AuthInfoContainer *i, KIO::AuthInfo& info)
 {
-    KIO::AuthInfo result = i->info;
-    result.setModified(true);
-
-    return result;
+    info = i->info;
+    info.setModified(true);
 }
 
 const KPasswdServer::AuthInfoContainer *
 KPasswdServer::findAuthInfoItem(const QString &key, const KIO::AuthInfo &info)
 {
-   kDebug(debugArea()) << "key:" << key << ", user=" << info.username;
+   // kDebug(debugArea()) << "key=" << key << ", user=" << info.username;
 
    AuthInfoContainerList *authList = m_authDict.value(key);
-   if (!authList)
-      return 0;
-
-   QString path2 = info.url.directory(KUrl::AppendTrailingSlash|KUrl::ObeyTrailingSlash);
-   Q_FOREACH(AuthInfoContainer *current, *authList)
+   if (authList)
    {
-       if ((current->expire == AuthInfoContainer::expTime) &&
-          (difftime(time(0), current->expireTime) > 0))
-       {
-          authList->removeOne(current);
-          delete current;
-          continue;
-       }
+      QString path2 = info.url.directory(KUrl::AppendTrailingSlash|KUrl::ObeyTrailingSlash);
+      Q_FOREACH(AuthInfoContainer *current, *authList)
+      {
+          if (current->expire == AuthInfoContainer::expTime &&
+              static_cast<qulonglong>(time(0)) > current->expireTime)
+          {
+              authList->removeOne(current);
+              delete current;
+              continue;
+          }
 
-       if (info.verifyPath)
-       {
-          QString path1 = current->directory;
-          if (path2.startsWith(path1) &&
-              (info.username.isEmpty() || info.username == current->info.username))
-             return current;
-       }
-       else
-       {
-          if (current->info.realmValue == info.realmValue &&
-              (info.username.isEmpty() || info.username == current->info.username))
-             return current; // TODO: Update directory info,
-       }
+          if (info.verifyPath)
+          {
+              QString path1 = current->directory;
+              if (path2.startsWith(path1) &&
+                  (info.username.isEmpty() || info.username == current->info.username))
+                return current;
+          }
+          else
+          {
+              if (current->info.realmValue == info.realmValue &&
+                  (info.username.isEmpty() || info.username == current->info.username))
+                return current; // TODO: Update directory info,
+          }
+      }
    }
    return 0;
 }
@@ -823,6 +699,13 @@ KPasswdServer::removeAuthInfoItem(const QString &key, const KIO::AuthInfo &info)
 void
 KPasswdServer::addAuthInfoItem(const QString &key, const KIO::AuthInfo &info, qlonglong windowId, qlonglong seqNr, bool canceled)
 {
+   kDebug(debugArea()) << "key=" << key
+                       << "window-id=" << windowId
+                       << "username=" << info.username
+                       << "realm=" << info.realmValue
+                       << "seqNr=" << seqNr
+                       << "keepPassword?" << info.keepPassword
+                       << "canceled?" << canceled;
    AuthInfoContainerList *authList = m_authDict.value(key);
    if (!authList)
    {
@@ -842,13 +725,9 @@ KPasswdServer::addAuthInfoItem(const QString &key, const KIO::AuthInfo &info, ql
 
    if (!authItem)
    {
+      kDebug(debugArea()) << "Creating AuthInfoContainer";
       authItem = new AuthInfoContainer;
       authItem->expire = AuthInfoContainer::expTime;
-      kDebug(debugArea()) << "Creating AuthInfoContainer";
-   }
-   else
-   {
-      kDebug(debugArea()) << "Updating AuthInfoContainer";
    }
 
    authItem->info = info;
@@ -867,7 +746,11 @@ void
 KPasswdServer::updateAuthExpire(const QString &key, const AuthInfoContainer *auth, qlonglong windowId, bool keep)
 {
    AuthInfoContainer *current = const_cast<AuthInfoContainer *>(auth);
-   if (keep)
+   Q_ASSERT(current);
+
+   kDebug(debugArea()) << "key=" << key << "expire=" << current->expire << "window-id=" << windowId << "keep=" << keep;
+
+   if (keep && !windowId)
    {
       current->expire = AuthInfoContainer::expNever;
    }
@@ -879,7 +762,7 @@ KPasswdServer::updateAuthExpire(const QString &key, const AuthInfoContainer *aut
    }
    else if (current->expire == AuthInfoContainer::expTime)
    {
-      current->expireTime = time(0)+10;
+      current->expireTime = time(0) + 10;
    }
 
    // Update mWindowIdList
@@ -901,18 +784,322 @@ KPasswdServer::removeAuthForWindowId(qlonglong windowId)
       if (!authList)
          continue;
 
-      Q_FOREACH(AuthInfoContainer *current, *authList)
+      QMutableListIterator<AuthInfoContainer*> it (*authList);
+      while (it.hasNext())
       {
+        AuthInfoContainer* current = it.next();
         if (current->expire == AuthInfoContainer::expWindowClose)
         {
            if (current->windowList.removeAll(windowId) && current->windowList.isEmpty())
            {
-              authList->removeOne(current);
               delete current;
+              it.remove();
            }
         }
       }
    }
 }
 
+void KPasswdServer::showPasswordDialog (KPasswdServer::Request* request)
+{
+    KIO::AuthInfo &info = request->info;
+    const bool bypassCacheAndKWallet = info.getExtraField(AUTHINFO_EXTRAFIELD_BYPASS_CACHE_AND_KWALLET).toBool();
+
+    QString username = info.username;
+    QString password = info.password;
+    bool hasWalletData = false;
+    QMap<QString, QString> knownLogins;
+
+    if ( !bypassCacheAndKWallet
+        && ( username.isEmpty() || password.isEmpty() )
+        && !KWallet::Wallet::keyDoesNotExist(KWallet::Wallet::NetworkWallet(), KWallet::Wallet::PasswordFolder(), makeWalletKey( request->key, info.realmValue )) )
+    {
+        // no login+pass provided, check if kwallet has one
+        if ( openWallet( request->windowId ) )
+            hasWalletData = readFromWallet( m_wallet, request->key, info.realmValue, username, password, info.readOnly, knownLogins );
+    }
+
+    // assemble dialog-flags
+    KPasswordDialog::KPasswordDialogFlags dialogFlags;
+
+    if (info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).isValid())
+    {
+        dialogFlags |= KPasswordDialog::ShowDomainLine;
+        if (info.getExtraFieldFlags(AUTHINFO_EXTRAFIELD_DOMAIN) & KIO::AuthInfo::ExtraFieldReadOnly)
+        {
+            dialogFlags |= KPasswordDialog::DomainReadOnly;
+        }
+    }
+
+    if (info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).isValid())
+    {
+        dialogFlags |= KPasswordDialog::ShowAnonymousLoginCheckBox;
+    }
+
+    if (!info.getExtraField(AUTHINFO_EXTRAFIELD_HIDE_USERNAME_INPUT).toBool())
+    {
+        dialogFlags |= KPasswordDialog::ShowUsernameLine;
+    }
+
+    // If wallet is not enabled and the caller explicitly requested for it,
+    // do not show the keep password checkbox.
+    if (info.keepPassword && KWallet::Wallet::isEnabled())
+        dialogFlags |= KPasswordDialog::ShowKeepPassword;
+
+    // instantiate dialog
+    kDebug(debugArea()) << "Widget for" << request->windowId << QWidget::find(request->windowId) << QApplication::activeWindow();
+    KPasswordDialog* dlg = new KPasswordDialog(0, dialogFlags);
+    connect(dlg, SIGNAL(finished(int)), this, SLOT(passwordDialogDone(int)));
+    connect(this, SIGNAL(destroyed(QObject*)), dlg, SLOT(deleteLater()));
+
+    dlg->setPrompt(info.prompt);
+    dlg->setUsername(username);
+    if (info.caption.isEmpty())
+        dlg->setPlainCaption( i18n("Authorization Dialog") );
+    else
+        dlg->setPlainCaption( info.caption );
+
+    if ( !info.comment.isEmpty() )
+        dlg->addCommentLine( info.commentLabel, info.comment );
+
+    if ( !password.isEmpty() )
+        dlg->setPassword( password );
+
+    if (info.readOnly)
+        dlg->setUsernameReadOnly( true );
+    else
+        dlg->setKnownLogins( knownLogins );
+
+    if (hasWalletData)
+        dlg->setKeepPassword( true );
+
+    if (info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).isValid ())
+        dlg->setDomain(info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).toString());
+
+    if (info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).isValid () && password.isEmpty() && username.isEmpty())
+        dlg->setAnonymousMode(info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).toBool());
+
+#ifndef Q_WS_WIN
+    KWindowSystem::setMainWindow(dlg, request->windowId);
+#else
+    KWindowSystem::setMainWindow(dlg, (HWND)(long)request->windowId);
+#endif
+
+    kDebug(debugArea()) << "Showing password dialog" << dlg << ", window-id=" << request->windowId;
+    m_authInProgress.insert(dlg, request);
+    dlg->open();
+}
+
+
+void KPasswdServer::sendResponse (KPasswdServer::Request* request)
+{
+    Q_ASSERT(request);
+    if (!request) {
+        return;
+    }
+
+    kDebug(debugArea()) << "key=" << request->key;
+    if (request->isAsync) {
+        emit queryAuthInfoAsyncResult(request->requestId, m_seqNr, request->info);
+    } else {
+        QByteArray replyData;
+        QDataStream stream2(&replyData, QIODevice::WriteOnly);
+        stream2 << request->info;
+        QDBusConnection::sessionBus().send(request->transaction.createReply(QVariantList() << replyData << m_seqNr));
+    }
+
+    // Check all requests in the wait queue.
+    Request *waitRequest;
+    QMutableListIterator<Request*> it(m_authWait);
+    while (it.hasNext()) {
+        waitRequest = it.next();
+
+        if (!hasPendingQuery(waitRequest->key, waitRequest->info))
+        {
+            const AuthInfoContainer *result = findAuthInfoItem(waitRequest->key,
+                                                               waitRequest->info);
+            QByteArray replyData;
+
+            QDataStream stream2(&replyData, QIODevice::WriteOnly);
+
+            KIO::AuthInfo rcinfo;
+            if (!result || result->isCanceled)
+            {
+                waitRequest->info.setModified(false);
+                stream2 << waitRequest->info;
+            }
+            else
+            {
+                updateAuthExpire(waitRequest->key, result, waitRequest->windowId, false);
+                copyAuthInfo(result, rcinfo);
+                stream2 << rcinfo;
+            }
+
+            if (waitRequest->isAsync) {
+                emit checkAuthInfoAsyncResult(waitRequest->requestId, m_seqNr, rcinfo);
+            } else {
+                QDBusConnection::sessionBus().send(waitRequest->transaction.createReply(QVariantList() << replyData << m_seqNr));
+            }
+
+            delete waitRequest;
+            it.remove();
+        }
+    }
+
+    // Re-enable password request processing for the current window id again.
+    m_authPrompted.removeAll(QString::number(request->windowId));
+    m_authPrompted.removeAll(request->key);
+
+    if (m_authPending.count())
+       QTimer::singleShot(0, this, SLOT(processRequest()));
+}
+
+void KPasswdServer::passwordDialogDone (int result)
+{
+    KPasswordDialog* dlg = qobject_cast<KPasswordDialog*>(sender());
+    Q_ASSERT(dlg);
+
+    QScopedPointer<Request> request (m_authInProgress.take(dlg));
+    Q_ASSERT(request);  // request should never be NULL.
+
+    if (request) {
+        KIO::AuthInfo& info = request->info;
+        const bool bypassCacheAndKWallet = info.getExtraField(AUTHINFO_EXTRAFIELD_BYPASS_CACHE_AND_KWALLET).toBool();
+
+        kDebug(debugArea()) << "dialog result=" << result << ", bypassCacheAndKWallet?" << bypassCacheAndKWallet;
+        if (dlg && result == KDialog::Accepted) {
+            Q_ASSERT(dlg);
+            const QString oldUsername (info.username);
+            info.username = dlg->username();
+            info.password = dlg->password();
+            info.keepPassword = dlg->keepPassword();
+
+            if (info.getExtraField(AUTHINFO_EXTRAFIELD_DOMAIN).isValid ())
+                info.setExtraField(AUTHINFO_EXTRAFIELD_DOMAIN, dlg->domain());
+            if (info.getExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS).isValid ())
+                info.setExtraField(AUTHINFO_EXTRAFIELD_ANONYMOUS, dlg->anonymousMode());
+
+            // When the user checks "keep password", that means:
+            // * if the wallet is enabled, store it there for long-term, and in kpasswdserver
+            // only for the duration of the window (#92928)
+            // * otherwise store in kpasswdserver for the duration of the KDE session.
+            if (!bypassCacheAndKWallet) {
+                /*
+                  NOTE: The following code changes the key under which the auth
+                  info is stored in memory if the request url contains a username.
+                  e.g. "ftp://user@localhost", but the user changes that username
+                  in the password dialog.
+
+                  Since the key generated to store the credential contains the
+                  username from the request URL, the key must be updated on such
+                  changes. Otherwise, the key will not be found on subsequent
+                  requests and the user will be end up being prompted over and
+                  over to re-enter the password unnecessarily.
+                */
+                if (!info.url.user().isEmpty() && info.username != info.url.user()) {
+                    const QString oldKey(request->key);
+                    removeAuthInfoItem(oldKey, info);
+                    info.url.setUser(info.username);
+                    request->key = createCacheKey(info);
+                    updateCachedRequestKey(m_authPending, oldKey, request->key);
+                    updateCachedRequestKey(m_authWait, oldKey, request->key);
+                }
+
+                const bool skipAutoCaching = info.getExtraField(AUTHINFO_EXTRAFIELD_SKIP_CACHING_ON_QUERY).toBool();
+                if (!skipAutoCaching && info.keepPassword && openWallet(request->windowId)) {
+                    if ( storeInWallet( m_wallet, request->key, info ) )
+                        // password is in wallet, don't keep it in memory after window is closed
+                        info.keepPassword = false;
+                }
+                addAuthInfoItem(request->key, info, request->windowId, m_seqNr, false);
+            }
+            info.setModified( true );
+        } else {
+            if (!bypassCacheAndKWallet && request->prompt) {
+                addAuthInfoItem(request->key, info, 0, m_seqNr, true);
+            }
+            info.setModified( false );
+        }
+
+        sendResponse(request.data());
+    }
+
+    dlg->deleteLater();
+}
+
+void KPasswdServer::retryDialogDone (int result)
+{
+    KDialog* dlg = qobject_cast<KDialog*>(sender());
+    Q_ASSERT(dlg);
+
+    QScopedPointer<Request> request (m_authRetryInProgress.take(dlg));
+    Q_ASSERT(request);
+
+    if (request) {
+        if (result == KDialog::Yes) {
+            showPasswordDialog(request.take());
+        } else {
+            // NOTE: If the user simply cancels the retry dialog, we remove the
+            // credential stored under this key because the original attempt to
+            // use it has failed. Otherwise, the failed credential would be cached
+            // and used subsequently.
+            //
+            // TODO: decide whether it should be removed from the wallet too.
+            KIO::AuthInfo& info = request->info;
+            removeAuthInfoItem(request->key, request->info);
+            info.setModified(false);
+            sendResponse(request.data());
+        }
+    }
+}
+
+void KPasswdServer::windowRemoved (WId id)
+{
+    bool foundMatch = false;
+    if (!m_authInProgress.isEmpty()) {
+        const qlonglong windowId = static_cast<qlonglong>(id);
+        QMutableHashIterator<QObject*, Request*> it (m_authInProgress);
+        while (it.hasNext()) {
+            it.next();
+            if (it.value()->windowId == windowId) {
+                Request* request = it.value();
+                QObject* obj = it.key();
+                it.remove();
+                m_authPrompted.removeAll(QString::number(request->windowId));
+                m_authPrompted.removeAll(request->key);
+                delete obj;
+                delete request;
+                foundMatch = true;
+            }
+        }
+    }
+
+    if (!foundMatch && !m_authRetryInProgress.isEmpty()) {
+        const qlonglong windowId = static_cast<qlonglong>(id);
+        QMutableHashIterator<QObject*, Request*> it (m_authRetryInProgress);
+        while (it.hasNext()) {
+            it.next();
+            if (it.value()->windowId == windowId) {
+                Request* request = it.value();
+                QObject* obj = it.key();
+                it.remove();
+                delete obj;
+                delete request;
+            }
+        }
+    }
+}
+
+void KPasswdServer::updateCachedRequestKey (QList<KPasswdServer::Request*>& list, const QString& oldKey, const QString& newKey)
+{
+    QListIterator<Request*> it (list);
+    while (it.hasNext()) {
+        Request* r = it.next();
+        if (r->key == oldKey) {
+            r->key = newKey;
+        }
+    }
+}
+
+
 #include "kpasswdserver.moc"
